from archinfo.types import RegisterOffset

from .arch import Arch, Endness, Register, register_arch
from .archerror import ArchError
from .tls import TLSArchInfo

try:
    import capstone as _capstone
except ImportError:
    _capstone = None

try:
    import keystone as _keystone
except ImportError:
    _keystone = None

try:
    import unicorn as _unicorn
except ImportError:
    _unicorn = None

try:
    import pyvex as _pyvex
except ImportError:
    _pyvex = None


_NATIVE_FUNCTION_PROLOGS = {
    rb"\x8b\xff\x55\x8b\xec",  # mov edi, edi; push ebp; mov ebp, esp
    rb"\x55\x8b\xec",  # push ebp; mov ebp, esp
    rb"\x55\x89\xe5",  # push ebp; mov ebp, esp
    rb"\x55\x57\x56",  # push ebp; push edi; push esi
    # mov eax, 0x000000??; (push ebp; push eax; push edi; push ebx; push esi; push edx; push ecx) sub esp
    rb"\xb8[\x00-\xff]\x00\x00\x00[\x50\x51\x52\x53\x55\x56\x57]{0,7}\x8b[\x00-\xff]{2}",
    # (push ebp; push eax; push edi; push ebx; push esi; push edx; push ecx) sub esp
    rb"[\x50\x51\x52\x53\x55\x56\x57]{1,7}\x83\xec[\x00-\xff]{2,4}",
    # (push ebp; push eax; push edi; push ebx; push esi; push edx; push ecx) mov xxx, xxx
    rb"[\x50\x51\x52\x53\x55\x56\x57]{1,7}\x8b[\x00-\xff]{2}",
    rb"(\x81|\x83)\xec",  # sub xxx %esp
}
# every function prolog can potentially be prefixed with endbr32
_endbr32 = b"\xf3\x0f\x1e\xfb"
_prefixed = {(_endbr32 + prolog) for prolog in _NATIVE_FUNCTION_PROLOGS}
_FUNCTION_PROLOGS = _prefixed | _NATIVE_FUNCTION_PROLOGS


class ArchX86(Arch):
    def __init__(self, endness=Endness.LE):
        if endness != Endness.LE:
            raise ArchError("Arch i386 must be little endian")
        super().__init__(endness)
        if self.vex_archinfo:
            self.vex_archinfo["x86_cr0"] = 0xFFFFFFFF

        # Register blacklist
        reg_blacklist = ("cs", "ds", "es", "fs", "gs", "ss", "gdt", "ldt")
        if self.reg_blacklist is not None and self.reg_blacklist_offsets is not None:
            for register in self.register_list:
                if register.name in reg_blacklist:
                    self.reg_blacklist.append(register.name)
                    self.reg_blacklist_offsets.append(register.vex_offset)

        if _unicorn and _pyvex:
            # CPU flag registers
            uc_flags_reg = _unicorn.x86_const.UC_X86_REG_EFLAGS
            cpu_flag_registers = {"d": 1 << 10, "ac": 1 << 18, "id": 1 << 21}
            for reg, reg_bitmask in cpu_flag_registers.items():
                reg_offset = self.get_register_offset(reg)
                self.cpu_flag_register_offsets_and_bitmasks_map[reg_offset] = (uc_flags_reg, reg_bitmask)

            mxcsr_registers = {"sseround": 1 << 14 | 1 << 13}
            uc_mxcsr_reg = _unicorn.x86_const.UC_X86_REG_MXCSR
            for reg, reg_bitmask in mxcsr_registers.items():
                reg_offset = self.get_register_offset(reg)
                self.cpu_flag_register_offsets_and_bitmasks_map[reg_offset] = (uc_mxcsr_reg, reg_bitmask)

    @property
    def capstone_x86_syntax(self):
        """
        Get the current syntax Capstone uses for x86. It can be 'intel' or 'at&t'

        :return: Capstone's current x86 syntax
        :rtype: str
        """

        return self._cs_x86_syntax

    @capstone_x86_syntax.setter
    def capstone_x86_syntax(self, new_syntax):
        """
        Set the syntax that Capstone outputs for x86.
        """

        if new_syntax not in ("intel", "at&t"):
            raise ArchError('Unsupported Capstone x86 syntax. It must be either "intel" or "at&t".')

        if new_syntax != self._cs_x86_syntax:
            self._cs = None
            self._cs_x86_syntax = new_syntax

    def _configure_capstone(self):
        self._cs.syntax = (
            _capstone.CS_OPT_SYNTAX_ATT if self._cs_x86_syntax == "at&t" else _capstone.CS_OPT_SYNTAX_INTEL
        )

    @property
    def keystone_x86_syntax(self):
        """
        Get the current syntax Keystone uses for x86. It can be 'intel',
        'at&t', 'nasm', 'masm', 'gas' or 'radix16'

        :return: Keystone's current x86 syntax
        :rtype: str
        """

        return self._ks_x86_syntax

    @keystone_x86_syntax.setter
    def keystone_x86_syntax(self, new_syntax):
        """
        Set the syntax that Keystone uses for x86.
        """

        if new_syntax not in ("intel", "at&t", "nasm", "masm", "gas", "radix16"):
            raise ArchError(
                "Unsupported Keystone x86 syntax. It must be one of the following: "
                '"intel", "at&t", "nasm", "masm", "gas" or "radix16".'
            )

        if new_syntax != self._ks_x86_syntax:
            self._ks = None
            self._ks_x86_syntax = new_syntax

    def _configure_keystone(self):
        if self._ks_x86_syntax == "at&t":
            self._ks.syntax = _keystone.KS_OPT_SYNTAX_ATT
        elif self._ks_x86_syntax == "nasm":
            self._ks.syntax = _keystone.KS_OPT_SYNTAX_NASM
        elif self._ks_x86_syntax == "masm":
            self._ks.syntax = _keystone.KS_OPT_SYNTAX_MASM
        elif self._ks_x86_syntax == "gas":
            self._ks.syntax = _keystone.KS_OPT_SYNTAX_GAS
        elif self._ks_x86_syntax == "radix16":
            self._ks.syntax = _keystone.KS_OPT_SYNTAX_RADIX16
        else:
            self._ks.syntax = _keystone.KS_OPT_SYNTAX_INTEL

    bits = 32
    vex_arch = "VexArchX86"
    name = "X86"
    qemu_name = "i386"
    ida_processor = "metapc"
    linux_name = "i386"
    triplet = "i386-linux-gnu"
    max_inst_bytes = 15
    call_sp_fix = -4
    ret_offset = RegisterOffset(8)
    vex_conditional_helpers = True
    syscall_num_offset = 8
    call_pushes_ret = True
    stack_change = -4
    memory_endness = Endness.LE
    register_endness = Endness.LE
    sizeof = {"short": 16, "int": 32, "long": 32, "long long": 64}
    if _capstone:
        cs_arch = _capstone.CS_ARCH_X86
        cs_mode = _capstone.CS_MODE_32 + _capstone.CS_MODE_LITTLE_ENDIAN
    _cs_x86_syntax = None  # Set it to 'att' in order to use AT&T syntax for x86
    if _keystone:
        ks_arch = _keystone.KS_ARCH_X86
        ks_mode = _keystone.KS_MODE_32 + _keystone.KS_MODE_LITTLE_ENDIAN
    _ks_x86_syntax = None
    uc_arch = _unicorn.UC_ARCH_X86 if _unicorn else None
    uc_mode = (_unicorn.UC_MODE_32 + _unicorn.UC_MODE_LITTLE_ENDIAN) if _unicorn else None
    uc_const = _unicorn.x86_const if _unicorn else None
    uc_prefix = "UC_X86_" if _unicorn else None
    function_prologs = _FUNCTION_PROLOGS
    function_epilogs = {
        rb"\xc9\xc3",  # leave; ret
        rb"([^\x41][\x50-\x5f]{1}|\x41[\x50-\x5f])\xc3",  # pop <reg>; ret
        rb"[^\x48][\x83,\x81]\xc4([\x00-\xff]{1}|[\x00-\xff]{4})\xc3",  #  add esp, <siz>; retq
    }
    ret_instruction = b"\xc3"
    nop_instruction = b"\x90"
    instruction_alignment = 1
    register_list = [
        Register(
            name="eax",
            size=4,
            subregisters=[("ax", 0, 2), ("al", 0, 1), ("ah", 1, 1)],
            general_purpose=True,
            argument=True,
            linux_entry_value=0x1C,
        ),
        Register(
            name="ecx",
            size=4,
            subregisters=[("cx", 0, 2), ("cl", 0, 1), ("ch", 1, 1)],
            general_purpose=True,
            argument=True,
        ),
        Register(
            name="edx",
            size=4,
            subregisters=[("dx", 0, 2), ("dl", 0, 1), ("dh", 1, 1)],
            general_purpose=True,
            argument=True,
            linux_entry_value="ld_destructor",
        ),
        Register(
            name="ebx",
            size=4,
            subregisters=[("bx", 0, 2), ("bl", 0, 1), ("bh", 1, 1)],
            general_purpose=True,
            argument=True,
        ),
        Register(
            name="esp",
            size=4,
            alias_names=("sp",),
            general_purpose=True,
            default_value=(Arch.initial_sp, True, "global"),
        ),
        Register(name="ebp", size=4, alias_names=("bp",), general_purpose=True, argument=True, linux_entry_value=0),
        Register(
            name="esi",
            size=4,
            subregisters=[("si", 0, 2), ("sil", 0, 1), ("sih", 1, 1)],
            general_purpose=True,
            argument=True,
        ),
        Register(
            name="edi",
            size=4,
            subregisters=[("di", 0, 2), ("dil", 0, 1), ("dih", 1, 1)],
            general_purpose=True,
            argument=True,
        ),
        Register(name="cc_op", size=4, default_value=(0, False, None), concrete=False, artificial=True),
        Register(name="cc_dep1", size=4, concrete=False, artificial=True),
        Register(name="cc_dep2", size=4, concrete=False, artificial=True),
        Register(name="cc_ndep", size=4, concrete=False, artificial=True),
        Register(name="d", size=4, alias_names=("dflag",), default_value=(1, False, None), concrete=False),
        Register(name="id", size=4, alias_names=("idflag",), default_value=(1, False, None), concrete=False),
        Register(name="ac", size=4, alias_names=("acflag",), default_value=(0, False, None), concrete=False),
        Register(name="eip", size=4, alias_names=("ip", "pc")),
        Register(
            name="fpreg",
            size=64,
            subregisters=[
                ("mm0", 0, 8),
                ("mm1", 8, 8),
                ("mm2", 16, 8),
                ("mm3", 24, 8),
                ("mm4", 32, 8),
                ("mm5", 40, 8),
                ("mm6", 48, 8),
                ("mm7", 56, 8),
            ],
            alias_names=("fpu_regs",),
            floating_point=True,
            concrete=False,
        ),
        Register(name="fptag", size=8, alias_names=("fpu_tags",), floating_point=True, default_value=(0, False, None)),
        Register(name="fpround", size=4, floating_point=True, default_value=(0, False, None)),
        Register(name="fc3210", size=4, floating_point=True),
        Register(name="ftop", size=4, floating_point=True, default_value=(7, False, None), artificial=True),
        Register(name="sseround", size=4, vector=True, default_value=(0, False, None)),
        Register(name="xmm0", size=16, vector=True),
        Register(name="xmm1", size=16, vector=True),
        Register(name="xmm2", size=16, vector=True),
        Register(name="xmm3", size=16, vector=True),
        Register(name="xmm4", size=16, vector=True),
        Register(name="xmm5", size=16, vector=True),
        Register(name="xmm6", size=16, vector=True),
        Register(name="xmm7", size=16, vector=True),
        Register(name="cs", size=2),
        Register(name="ds", size=2),
        Register(name="es", size=2),
        Register(name="fs", size=2, default_value=(0, False, None), concrete=False),
        Register(name="gs", size=2, default_value=(0, False, None), concrete=False),
        Register(name="ss", size=2),
        Register(name="ldt", size=8, default_value=(0, False, None), concrete=False),
        Register(name="gdt", size=8, default_value=(0, False, None), concrete=False),
        Register(name="emnote", size=4, artificial=True),
        Register(name="cmstart", size=4),
        Register(name="cmlen", size=4),
        Register(name="nraddr", size=4),
        Register(name="sc_class", size=4),
        Register(name="ip_at_syscall", size=4, concrete=False, artificial=True),
    ]

    symbol_type_translation = {10: "STT_GNU_IFUNC", "STT_LOOS": "STT_GNU_IFUNC"}
    lib_paths = ["/lib32", "/usr/lib32"]
    got_section_name = ".got.plt"
    ld_linux_name = "ld-linux.so.2"
    elf_tls = TLSArchInfo(2, 56, [8], [4], [0], 0, 0)
    dwarf_registers = [
        "eax",
        "ecx",
        "edx",
        "ebx",
        "esp",
        "ebp",
        "esi",
        "edi",
        "eip",
        "eflags",
        "<none>",
        "st0",
        "st1",
        "st2",
        "st3",
        "st4",
        "st5",
        "st6",
        "st7",
        "<none>",
        "<none>",
        "xmm0",
        "xmm1",
        "xmm2",
        "xmm3",
        "xmm4",
        "xmm5",
        "xmm6",
        "xmm7",
        "mm0",
        "mm1",
        "mm2",
        "mm3",
        "mm4",
        "mm5",
        "mm6",
        "mm7",
        "fcw",
        "fsw",
        "mxcsr",
        "es",
        "cs",
        "ss",
        "ds",
        "fs",
        "gs",
        "<none>",
        "<none>",
        "tr",
        "ldtr",
    ]


register_arch([r".*i?\d86|.*x32|.*x86|.*metapc"], 32, Endness.LE, ArchX86)



import unittest

from archinfo import ArchAMD64, ArchError
from archinfo.arch import Endness


class TestAmd64(unittest.TestCase):
    def test_arch_amd64(self):
        endness = Endness.LE
        assert ArchAMD64(endness)

    def test_arch_amd64_passes(self):
        with self.assertRaises(ArchError):
            endness = Endness.BE
            ArchAMD64(endness)

    def test_capstone_x86_syntax(self):
        inst_1 = ArchAMD64(endness=Endness.LE)
        assert inst_1.capstone_x86_syntax is None
        inst_1.capstone_x86_syntax = "intel"
        assert inst_1.capstone_x86_syntax == "intel"
        inst_1.capstone_x86_syntax = "at&t"
        assert inst_1.capstone_x86_syntax == "at&t"

    # Test raises one of expected exceptions to pass.
    def test_capstone_x86_syntax_fails_1(self):
        with self.assertRaises(ArchError):
            inst_1 = ArchAMD64(endness=Endness.LE)
            inst_1.capstone_x86_syntax = "at&"
            assert inst_1.capstone_x86_syntax

    def test_capstone_x86_syntax_fails_2(self):
        with self.assertRaises(ArchError):
            inst_1 = ArchAMD64(endness=Endness.LE)
            inst_1.capstone_x86_syntax = "int"
            assert inst_1.capstone_x86_syntax

    def test_keystone_x86_syntax(self):
        inst_1 = ArchAMD64(endness=Endness.LE)
        assert inst_1.keystone_x86_syntax is None
        inst_1.keystone_x86_syntax = "intel"
        assert inst_1.keystone_x86_syntax == "intel"
        inst_1.keystone_x86_syntax = "at&t"
        assert inst_1.keystone_x86_syntax == "at&t"
        inst_1.keystone_x86_syntax = "nasm"
        assert inst_1.keystone_x86_syntax == "nasm"
        inst_1.keystone_x86_syntax = "masm"
        assert inst_1.keystone_x86_syntax == "masm"
        inst_1.keystone_x86_syntax = "gas"
        assert inst_1.keystone_x86_syntax == "gas"
        inst_1.keystone_x86_syntax = "radix16"
        assert inst_1.keystone_x86_syntax == "radix16"

    # Test raises one of expected exceptions to pass.
    def test_keystone_x86_syntax_fails_1(self):
        with self.assertRaises(ArchError):
            inst_1 = ArchAMD64(endness=Endness.LE)
            inst_1.keystone_x86_syntax = "inte"
            assert inst_1.keystone_x86_syntax

    def test_keystone_x86_syntax_fails_2(self):
        with self.assertRaises(ArchError):
            inst_1 = ArchAMD64(endness=Endness.LE)
            inst_1.keystone_x86_syntax = "at"
            assert inst_1.keystone_x86_syntax

    def test_keystone_x86_syntax_fails_3(self):
        with self.assertRaises(ArchError):
            inst_1 = ArchAMD64(endness=Endness.LE)
            inst_1.keystone_x86_syntax = "na"
            assert inst_1.keystone_x86_syntax

    def test_keystone_x86_syntax_fails_4(self):
        with self.assertRaises(ArchError):
            inst_1 = ArchAMD64(endness=Endness.LE)
            inst_1.keystone_x86_syntax = "ma"
            assert inst_1.keystone_x86_syntax



import logging
import re

from .arch import Arch, Endness, register_arch

log = logging.getLogger("archinfo.arch_soot")


class SootMethodDescriptor:
    __slots__ = ["class_name", "name", "params", "_soot_method", "ret"]

    def __init__(self, class_name, name, params, soot_method=None, ret_type=None):
        self.class_name = class_name
        self.name = name
        self.params = params
        self._soot_method = soot_method
        self.ret = ret_type

    def __repr__(self):
        return "{}.{}({})".format(self.class_name, self.name, ", ".join(self.params))

    def __hash__(self):
        return hash((self.class_name, self.name, self.params))

    def __eq__(self, other):
        return (
            isinstance(other, SootMethodDescriptor)
            and self.class_name == other.class_name
            and self.name == other.name
            and self.params == other.params
        )

    def __ne__(self, other):
        return not self == other

    def __lt__(self, other):
        return self.__repr__() < other.__repr__()

    def __gt__(self, other):
        return self.__repr__() > other.__repr__()

    def __le__(self, other):
        return self.__repr__() <= other.__repr__()

    def __ge__(self, other):
        return self.__repr__() >= other.__repr__()

    def address(self, block_idx=0, stmt_idx=0):
        """
        :return Address of the method.
        :rtype: SootAddressDescriptor
        """
        return SootAddressDescriptor(self, block_idx, stmt_idx)

    @property
    def fullname(self):
        """
        :return the full name of the method (class name + method name)
        """
        return f"{self.class_name}.{self.name}"

    @property
    def symbolic(self):
        return False

    @property
    def is_loaded(self):
        """
        :return: True, if the method is loaded in CLE and thus infos about attrs,
                 ret and exceptions are available.
        """
        return self._soot_method is not None

    @property
    def attrs(self):
        return self._soot_method.attrs if self.is_loaded else []

    @property
    def exceptions(self):
        return self._soot_method.exceptions if self.is_loaded else []

    @property
    def block_by_label(self):
        return self._soot_method.block_by_label if self.is_loaded else None

    # @property
    # def ret(self):
    #     return self._soot_method.ret if self.is_loaded else []

    @property
    def addr(self):
        """
        :return: the soot address description of the entry point of the method
        """
        return SootAddressDescriptor(self, 0, 0)

    def matches_with_native_name(self, native_method):
        """
        The name of native methods are getting encoded, s.t. they translate into
        valid C function names. This method indicates if the name of the given
        native method matches the name of the soot method.

        :return: True, if name of soot method matches the mangled native name.
        """

        if "__" in native_method:
            # if native methods are overloaded, two underscores are used
            native_method, params_sig = native_method.split("__")
            params = ArchSoot.decode_parameter_list_signature(params_sig)
            # check function signature
            if params != self.params:
                return False

        # demangle native name
        native_method = native_method.replace("_1", "_")
        # TODO unicode escaping

        method_native_name = "Java_{class_name}_{method_name}".format(
            class_name=self.class_name.replace(".", "_"), method_name=self.name
        )

        return native_method == method_native_name

    @classmethod
    def from_string(cls, tstr):
        # this should be the opposite of repr
        tstr = tstr.strip()
        class_and_method, tparams = tstr.split("(")
        params_str = tparams.split(")")[0]
        if params_str == "":
            params = ()
        else:
            params = tuple(t.strip() for t in params_str.split(","))
        class_name, _, method = class_and_method.rpartition(".")
        return cls(class_name, method, params)

    @classmethod
    def from_soot_method(cls, soot_method):
        return cls(
            class_name=soot_method.class_name,
            name=soot_method.name,
            params=soot_method.params,
            soot_method=soot_method,
            ret_type=soot_method.ret,
        )


class SootAddressDescriptor:
    __slots__ = ["method", "block_idx", "stmt_idx"]

    def __init__(self, method, block_idx, stmt_idx):
        if not isinstance(method, SootMethodDescriptor):
            raise ValueError('The parameter "method" must be an ' "instance of SootMethodDescriptor.")

        self.method = method
        self.block_idx = block_idx
        self.stmt_idx = stmt_idx

    def __repr__(self):
        return "<{!r}+({}:{})>".format(
            self.method, self.block_idx, "%d" % self.stmt_idx if self.stmt_idx is not None else "[0]"
        )

    def __hash__(self):
        return hash((self.method, self.stmt_idx))

    def __eq__(self, other):
        return (
            isinstance(other, SootAddressDescriptor)
            and self.method == other.method
            and self.block_idx == other.block_idx
            and self.stmt_idx == other.stmt_idx
        )

    def __ne__(self, other):
        return not self == other

    def __lt__(self, other):
        return self.__repr__() < other.__repr__()

    def __gt__(self, other):
        return self.__repr__() > other.__repr__()

    def __le__(self, other):
        return self.__repr__() <= other.__repr__()

    def __ge__(self, other):
        return self.__repr__() >= other.__repr__()

    def __add__(self, stmts_offset):
        if not isinstance(stmts_offset, int):
            raise TypeError("The stmts_offset must be an int or a long.")
        s = self.copy()
        s.stmt_idx += stmts_offset
        return s

    def copy(self):
        return SootAddressDescriptor(method=self.method, block_idx=self.block_idx, stmt_idx=self.stmt_idx)

    @property
    def symbolic(self):
        return False


class SootAddressTerminator(SootAddressDescriptor):
    __slots__ = []

    def __init__(self):
        dummy_method = SootMethodDescriptor("dummy", "dummy", tuple())
        super().__init__(dummy_method, 0, 0)

    def __repr__(self):
        return "<Terminator>"


class SootFieldDescriptor:
    __slots__ = ["class_name", "name", "type"]

    def __init__(self, class_name, name, type_):
        self.class_name = class_name
        self.name = name
        self.type = type_

    def __repr__(self):
        return f"{self.class_name}.{self.name}"

    def __hash__(self):
        return hash((self.class_name, self.name, self.type))

    def __eq__(self, other):
        return (
            isinstance(other, SootFieldDescriptor)
            and self.class_name == other.class_name
            and self.name == other.name
            and self.type == other.type
        )

    def __ne__(self, other):
        return not self == other


class SootClassDescriptor:
    __slots__ = ["name", "_soot_class"]

    def __init__(self, name, soot_class=None):
        self.name = name
        self._soot_class = soot_class

    def __repr__(self):
        return self.name

    def __hash__(self):
        return hash(self.name)

    def __eq__(self, other):
        return isinstance(other, SootClassDescriptor) and self.name == other.name

    def __ne__(self, other):
        return not self == other

    @property
    def is_loaded(self):
        """
        :return: True, if the class is loaded in CLE and thus info about field,
                 methods, ... are available.
        """
        return self._soot_class is not None

    @property
    def fields(self):
        return self._soot_class.fields if self.is_loaded else None

    @property
    def methods(self):
        return self._soot_class.methods if self.is_loaded else None

    @property
    def superclass_name(self):
        return self._soot_class.super_class if self.is_loaded else None

    @property
    def type(self):
        return "java.lang.Class"


class SootNullConstant:
    def __init__(self):
        pass

    def __repr__(self):
        return "null"

    def __hash__(self):
        return hash("null")

    def __eq__(self, other):
        return isinstance(other, SootNullConstant)

    def __ne__(self, other):
        return not self == other


class SootArgument:
    """
    Typed Java argument.
    """

    __slots__ = ["value", "type", "is_this_ref"]

    def __init__(self, value, type_, is_this_ref=False):
        """
        :param value:        Value of the argument
        :param type_:        Type of the argument
        :param is_this_ref:  Indicates whether the argument represents the
                             'this' reference, i.e. the object on which an
                             instance method is invoked.
        """
        self.value = value
        self.type = type_
        self.is_this_ref = is_this_ref

    def __repr__(self):
        return f"{self.value} ({self.type})"


class ArchSoot(Arch):
    def __init__(self, endness=Endness.LE):
        super().__init__(endness)

    name = "Soot"

    vex_arch = None  # No VEX support
    qemu_name = None  # No Qemu/Unicorn-engine support
    bits = 64
    address_types = (SootAddressDescriptor,)
    function_address_types = (SootMethodDescriptor,)

    # Size of native counterparts of primitive Java types
    sizeof = {"boolean": 8, "byte": 8, "char": 16, "short": 16, "int": 32, "long": 64, "float": 32, "double": 64}

    primitive_types = ["boolean", "byte", "char", "short", "int", "long", "float", "double"]

    sig_dict = {
        "Z": "boolean",
        "B": "byte",
        "C": "char",
        "S": "short",
        "I": "int",
        "J": "long",
        "F": "float",
        "D": "double",
        "V": "void",
    }

    @staticmethod
    def decode_type_signature(type_sig):
        if not type_sig:
            return None
        # try to translate signature as a primitive type
        if type_sig in ArchSoot.sig_dict:
            return ArchSoot.sig_dict[type_sig]
        # java classes are encoded as 'Lclass_name;'
        if type_sig.startswith("L") and type_sig.endswith(";"):
            return type_sig[1:-1]
        raise ValueError(type_sig)

    @staticmethod
    def decode_parameter_list_signature(param_sig):
        return tuple(
            ArchSoot.decode_type_signature(param) for param in re.findall(r"([\[]*[ZBCSIJFDV]|[\[]*L.+;)", param_sig)
        )

    @staticmethod
    def decode_method_signature(method_sig):
        # signature format follows the pattern: (param_sig)ret_sig
        match = re.search(r"\((.*)\)(.*)", method_sig)
        param_sig, ret_sig = match.group(1), match.group(2)
        # decode types
        params_types = ArchSoot.decode_parameter_list_signature(param_sig)
        ret_type = ArchSoot.decode_type_signature(ret_sig)
        log.debug("Decoded method signature '%s' as params=%s and ret=%s", method_sig, params_types, ret_type)
        return params_types, ret_type

    def library_search_path(self, pedantic=False):
        """
        Since Java is mostly system independent, we cannot return system
        specific paths.

        :return: empty list
        """
        return []


register_arch(["soot"], 8, Endness.LE, ArchSoot)



def test_register():
    from archinfo.arch import Register

    register = Register(
        name="MIPS",
        size=8,
        vex_offset=10,
        vex_name="xyz",
        subregisters=["a", "b"],
        alias_names=("r0"),
        general_purpose=True,
        floating_point=True,
        vector=True,
        argument=True,
        persistent=True,
        default_value=("a", "global"),
        linux_entry_value="argv",
        concretize_unique=True,
        concrete=True,
        artificial=True,
    )
    assert register.name == "MIPS"
    assert register.size == 8
    assert register.vex_offset == 10
    assert register.vex_name == "xyz"
    assert register.subregisters == ["a", "b"]
    assert register.alias_names == ("r0")
    assert register.general_purpose
    assert register.floating_point
    assert register.vector
    assert register.argument
    assert register.persistent
    assert register.default_value == ("a", "global")
    assert register.linux_entry_value == "argv"
    assert register.concretize_unique
    assert register.concrete
    assert register.artificial



import sys
from typing import NewType

if sys.version_info < (3, 11):
    from backports.strenum import StrEnum
else:
    from enum import StrEnum


class RegisterOffset(int):
    """A register offset is an integer that represents the offset of a register
    in VEX's register memory space.
    """

    def __add__(self, other):
        if isinstance(other, int):
            return RegisterOffset(int(self) + other)
        return NotImplemented


TmpVar = NewType("TmpVar", int)

# This causes too much issues as a NewType, sot is a simple alias instead
# This means that is still legal to pass any str where a RegisterName is expected.
# The downside is that PyCharm will show the type as `str` when displaying the signature
RegisterName = str


class Endness(StrEnum):
    """Endness specifies the byte order for integer values

    :cvar LE:      little endian, least significant byte is stored at lowest address
    :cvar BE:      big endian, most significant byte is stored at lowest address
    :cvar ME:      Middle-endian. Yep.
    """

    LE = "Iend_LE"
    BE = "Iend_BE"
    ME = "Iend_ME"
    ANY = "any"
    UNSURE = "unsure"

    @staticmethod
    def from_str(s: str) -> "Endness":
        for e in Endness:
            if e.value == s:
                return e
        raise ValueError("Unknown endness: %s" % s)



# pylint:disable=missing-class-docstring,no-self-use
import pickle
import unittest

try:
    import pypcode
except ImportError:
    pypcode = None

from archinfo import ArchError, ArchPcode
from archinfo.arch import Endness


@unittest.skipUnless(pypcode is not None, "pypcode not installed")
class TestArchPcode(unittest.TestCase):
    def test_arch_68000(self):
        arch = ArchPcode("68000:BE:32:default")
        assert arch.instruction_endness == Endness.BE
        assert arch.bits == 32

    def test_arch_bad_langid(self):
        with self.assertRaises(ArchError):
            ArchPcode("invalid")

    def test_pickle(self):
        arch = ArchPcode("68000:BE:32:default")
        pickle.dumps(arch)


if __name__ == "__main__":
    unittest.main()
